(* MD5 Message-Digest (rfc1321) *)
IMPLEMENTATION MODULE MD5;

FROM SYSTEM IMPORT BYTE, ADR, CAST;
IMPORT Types;

IMPORT Lib;

TYPE
  TCnvDesc = RECORD
    CASE : CARDINAL OF
      | 0: b: BYTE;
      | 1: c: CHAR;
      | 2: n: CARDINAL;
    END;
  END;
  (* access to TBufDesc as LONGCARDs *)
  TBuf = POINTER TO ARRAY [ 0 .. 15 ] OF Types.TLngCard;

  A17 = ARRAY [0..16] OF BYTE;


CONST
  bitsAll = Types.TLngSet{ 0 .. 31 };


<* PUSH *> <* COVERFLOW - *>
PROCEDURE RotlL( x: Types.TLngCard; n: CARDINAL ): Types.TLngCard;
BEGIN
  RETURN x  *  Types.TLngCard( Types.TLngSet{ n      } ) +
         x DIV Types.TLngCard( Types.TLngSet{ 32 - n } );
END RotlL;
<* POP *>


(* implicit BYTE length = SIZE(TBufDesc) *)
PROCEDURE Transform( VAR st: TStateDesc; bp: TBuf );

  PROCEDURE F( x, y, z: Types.TLngSet ): Types.TLngCard;
  BEGIN
    RETURN Types.TLngCard( x * y + ( x / bitsAll ) * z );
  END F;

  PROCEDURE G( x, y, z: Types.TLngSet ): Types.TLngCard;
  BEGIN
    RETURN Types.TLngCard( x * z + y * ( z / bitsAll ) );
  END G;

  PROCEDURE H( x, y, z: Types.TLngSet ): Types.TLngCard;
  BEGIN
    RETURN Types.TLngCard( x / y / z );
  END H;

  PROCEDURE I( x, y, z: Types.TLngSet ): Types.TLngCard;
  BEGIN
    RETURN Types.TLngCard( y / ( x + ( z / bitsAll ) ) );
  END I;

  <* PUSH *> <* COVERFLOW - *>
  PROCEDURE FF( VAR a: Types.TLngSet; b, c, d: Types.TLngSet; x: Types.TLngCard; s: CARDINAL );
  BEGIN
    a := Types.TLngSet( Types.TLngCard(b) + RotlL( Types.TLngCard(a) + F(b,c,d) + x, s ) );
  END FF;

  PROCEDURE GG( VAR a: Types.TLngSet; b, c, d: Types.TLngSet; x: Types.TLngCard; s: CARDINAL );
  BEGIN
    a := Types.TLngSet( Types.TLngCard(b) + RotlL( Types.TLngCard(a) + G(b,c,d) + x, s ) );
  END GG;

  PROCEDURE HH( VAR a: Types.TLngSet; b, c, d: Types.TLngSet; x: Types.TLngCard; s: CARDINAL );
  BEGIN
    a := Types.TLngSet( Types.TLngCard(b) + RotlL( Types.TLngCard(a) + H(b,c,d) + x, s ) );
  END HH;

  PROCEDURE II( VAR a: Types.TLngSet; b, c, d: Types.TLngSet; x: Types.TLngCard; s: CARDINAL );
  BEGIN
    a := Types.TLngSet( Types.TLngCard(b) + RotlL( Types.TLngCard(a) + I(b,c,d) + x, s ) );
  END II;
  <* POP *>

CONST
  S11 =  7;
  S12 = 12;
  S13 = 17;
  S14 = 22;
  S21 =  5;
  S22 =  9;
  S23 = 14;
  S24 = 20;
  S31 =  4;
  S32 = 11;
  S33 = 16;
  S34 = 23;
  S41 =  6;
  S42 = 10;
  S43 = 15;
  S44 = 21;
VAR
  a, b, c, d: Types.TLngSet;
BEGIN
<* PUSH *> <* COVERFLOW - *>
  a := st.a;
  b := st.b;
  c := st.c;
  d := st.d;

  (* Round 1 *)
  FF( a, b, c, d, bp^[ 0] + 0D76AA478H, S11 );
  FF( d, a, b, c, bp^[ 1] + 0E8C7B756H, S12 );
  FF( c, d, a, b, bp^[ 2] + 0242070DBH, S13 );
  FF( b, c, d, a, bp^[ 3] + 0C1BDCEEEH, S14 );
  FF( a, b, c, d, bp^[ 4] + 0F57C0FAFH, S11 );
  FF( d, a, b, c, bp^[ 5] + 04787C62AH, S12 );
  FF( c, d, a, b, bp^[ 6] + 0A8304613H, S13 );
  FF( b, c, d, a, bp^[ 7] + 0FD469501H, S14 );
  FF( a, b, c, d, bp^[ 8] + 0698098D8H, S11 );
  FF( d, a, b, c, bp^[ 9] + 08B44F7AFH, S12 );
  FF( c, d, a, b, bp^[10] + 0FFFF5BB1H, S13 );
  FF( b, c, d, a, bp^[11] + 0895CD7BEH, S14 );
  FF( a, b, c, d, bp^[12] + 06B901122H, S11 );
  FF( d, a, b, c, bp^[13] + 0FD987193H, S12 );
  FF( c, d, a, b, bp^[14] + 0A679438EH, S13 );
  FF( b, c, d, a, bp^[15] + 049B40821H, S14 );

  (* Round 2 *)
  GG( a, b, c, d, bp^[ 1] + 0F61E2562H, S21 );
  GG( d, a, b, c, bp^[ 6] + 0C040B340H, S22 );
  GG( c, d, a, b, bp^[11] + 0265E5A51H, S23 );
  GG( b, c, d, a, bp^[ 0] + 0E9B6C7AAH, S24 );
  GG( a, b, c, d, bp^[ 5] + 0D62F105DH, S21 );
  GG( d, a, b, c, bp^[10] + 002441453H, S22 );
  GG( c, d, a, b, bp^[15] + 0D8A1E681H, S23 );
  GG( b, c, d, a, bp^[ 4] + 0E7D3FBC8H, S24 );
  GG( a, b, c, d, bp^[ 9] + 021E1CDE6H, S21 );
  GG( d, a, b, c, bp^[14] + 0C33707D6H, S22 );
  GG( c, d, a, b, bp^[ 3] + 0F4D50D87H, S23 );
  GG( b, c, d, a, bp^[ 8] + 0455A14EDH, S24 );
  GG( a, b, c, d, bp^[13] + 0A9E3E905H, S21 );
  GG( d, a, b, c, bp^[ 2] + 0FCEFA3F8H, S22 );
  GG( c, d, a, b, bp^[ 7] + 0676F02D9H, S23 );
  GG( b, c, d, a, bp^[12] + 08D2A4C8AH, S24 );

  (* Round 3 *)
  HH( a, b, c, d, bp^[ 5] + 0FFFA3942H, S31 );
  HH( d, a, b, c, bp^[ 8] + 08771F681H, S32 );
  HH( c, d, a, b, bp^[11] + 06D9D6122H, S33 );
  HH( b, c, d, a, bp^[14] + 0FDE5380CH, S34 );
  HH( a, b, c, d, bp^[ 1] + 0A4BEEA44H, S31 );
  HH( d, a, b, c, bp^[ 4] + 04BDECFA9H, S32 );
  HH( c, d, a, b, bp^[ 7] + 0F6BB4B60H, S33 );
  HH( b, c, d, a, bp^[10] + 0BEBFBC70H, S34 );
  HH( a, b, c, d, bp^[13] + 0289B7EC6H, S31 );
  HH( d, a, b, c, bp^[ 0] + 0EAA127FAH, S32 );
  HH( c, d, a, b, bp^[ 3] + 0D4EF3085H, S33 );
  HH( b, c, d, a, bp^[ 6] + 004881D05H, S34 );
  HH( a, b, c, d, bp^[ 9] + 0D9D4D039H, S31 );
  HH( d, a, b, c, bp^[12] + 0E6DB99E5H, S32 );
  HH( c, d, a, b, bp^[15] + 01FA27CF8H, S33 );
  HH( b, c, d, a, bp^[ 2] + 0C4AC5665H, S34 );

  (* Round 4 *)
  II( a, b, c, d, bp^[ 0] + 0F4292244H, S41 );
  II( d, a, b, c, bp^[ 7] + 0432AFF97H, S42 );
  II( c, d, a, b, bp^[14] + 0AB9423A7H, S43 );
  II( b, c, d, a, bp^[ 5] + 0FC93A039H, S44 );
  II( a, b, c, d, bp^[12] + 0655B59C3H, S41 );
  II( d, a, b, c, bp^[ 3] + 08F0CCC92H, S42 );
  II( c, d, a, b, bp^[10] + 0FFEFF47DH, S43 );
  II( b, c, d, a, bp^[ 1] + 085845DD1H, S44 );
  II( a, b, c, d, bp^[ 8] + 06FA87E4FH, S41 );
  II( d, a, b, c, bp^[15] + 0FE2CE6E0H, S42 );
  II( c, d, a, b, bp^[ 6] + 0A3014314H, S43 );
  II( b, c, d, a, bp^[13] + 04E0811A1H, S44 );
  II( a, b, c, d, bp^[ 4] + 0F7537E82H, S41 );
  II( d, a, b, c, bp^[11] + 0BD3AF235H, S42 );
  II( c, d, a, b, bp^[ 2] + 02AD7D2BBH, S43 );
  II( b, c, d, a, bp^[ 9] + 0EB86D391H, S44 );

  st.a := Types.TLngSet( Types.TLngCard(st.a) + Types.TLngCard(a) );
  st.b := Types.TLngSet( Types.TLngCard(st.b) + Types.TLngCard(b) );
  st.c := Types.TLngSet( Types.TLngCard(st.c) + Types.TLngCard(c) );
  st.d := Types.TLngSet( Types.TLngCard(st.d) + Types.TLngCard(d) );
<* POP *>
END Transform;

<* PUSH *> <* COVERFLOW - *>
PROCEDURE IncCnt( VAR cnt: TCntDesc; x: Types.TLngCard );
BEGIN
  INC( cnt.l, x );
  IF cnt.l < x THEN INC( cnt.h ) END;
END IncCnt;
<* POP *>

PROCEDURE BufIdx( VAR cnt: TCntDesc ): TBufIdx;
BEGIN
  RETURN TBufIdx( ( cnt.l DIV 8 ) MOD SIZE(TBufDesc) );
END BufIdx;


(********** interface **********)


PROCEDURE Init( VAR c: TCtxDesc );
BEGIN
  c.st.a := CAST( Types.TLngSet, 067452301H );
  c.st.b := CAST( Types.TLngSet, 0EFCDAB89H );
  c.st.c := CAST( Types.TLngSet, 098BADCFEH );
  c.st.d := CAST( Types.TLngSet, 010325476H );
  c.cnt.l := 0;
  c.cnt.h := 0;
END Init;


PROCEDURE Update( VAR ctx: TCtxDesc; x-: ARRAY OF BYTE; len: CARDINAL );
VAR
  idx: TBufIdx;       (* 1st unused buffer element *)
  free: CARDINAL;     (* number of unused buffer elements *)
  n: CARDINAL;        (* number of bytes processed *)
BEGIN
  idx := BufIdx( ctx.cnt );
  IncCnt( ctx.cnt, 8 * Types.TLngCard( len ) );
  free := SIZE(TBufDesc) - idx;
  IF len >= free THEN
    (* transform, buffer remaining part only *)
    (* first part via buf *)
    Lib.Move( ADR(x), ADR(ctx.buf[idx]), free );
    Transform( ctx.st, ADR(ctx.buf) );
    idx := 0;
    (* middle part(s) directly *)
    n := free;
    WHILE n + SIZE(TBufDesc) <= len DO
      Transform( ctx.st, ADR(x[n]) );
      INC( n, SIZE(TBufDesc) );
    END;
  ELSE
    (* don't transform, buffer whole input *)
    n := 0;
  END;
  (* buffer remaining part *)
  IF len > n THEN
    Lib.Move( ADR(x[n]), ADR(ctx.buf[idx]), len-n );
  END;
END Update;


PROCEDURE Final( VAR ctx: TCtxDesc; VAR dgst: TDgstDesc );
VAR
  cnt: TCntDesc;   (* backup *)
  idx: TBufIdx;    (* 1st unused buffer element *)
  pad: TBufDesc;   (* padding data *)
  len: CARDINAL;   (* padding length *)
BEGIN
  cnt := ctx.cnt;
  Lib.Fill( ADR(pad), SIZE(pad), 0C );
  pad[ MIN(TBufIdx) ] := BYTE( CHR(080H) );
  idx := BufIdx( ctx.cnt );
  IF idx < 56 THEN
    len := 56 - idx;
  ELSE
    len := 120 - idx;
  END;
  Update( ctx, pad, len );
  Update( ctx, cnt, SIZE(cnt) );
  Lib.Move( ADR(ctx.st), ADR(dgst), SIZE(dgst) );
END Final;


PROCEDURE MkStr( VAR s: TDgstStr; d-: TDgstDesc );

  PROCEDURE Char( x: CARDINAL ): CHAR;
  BEGIN
    CASE x OF
    | 00H..09H: RETURN CHR( x + ORD( '0' ) );
    | 0AH..0FH: RETURN CHR( x + ORD( 'A' )  - 10 );
    ELSE
      RETURN '?';
    END;
  END Char;

VAR
  i,j: CARDINAL;
  cnv: TCnvDesc;
BEGIN
  cnv.n := 0;
  i := MIN( TDgstStrIdx );
  FOR j := MIN(TDgstIdx) TO MAX(TDgstIdx) DO
    cnv.b := d[ j ];
    s[ i ] := Char( cnv.n DIV 10H );
    INC( i );
    s[ i ] := Char( cnv.n MOD 10H );
    INC( i );
  END;
END MkStr;


PROCEDURE MkDgst( VAR d: TDgstDesc; s-: ARRAY OF CHAR ): BOOLEAN;
VAR
  i,j: CARDINAL;
  cnv: TCnvDesc;
  c: CHAR;
  x: CARDINAL;
BEGIN
  IF SIZE(s) < SIZE(TDgstStr) THEN RETURN FALSE END;
  j := MIN( TDgstIdx );
  FOR i := 0 TO SIZE(TDgstStr) - 1 DO
    c := CAP( s[i] );
    CASE c OF
    | '0'..'9': x := ORD( c ) - ORD( '0' );
    | 'A'..'F': x := ORD( c ) - ORD( 'A' ) + 10;
    ELSE
      RETURN FALSE;
    END;
    IF ODD( i ) THEN
      cnv.n := cnv.n + x;  (* lsd *)
      d[ j ] := cnv.b;
      INC( j );
    ELSE
      cnv.n := x * 10H;    (* msd *)
    END;
  END;
  RETURN TRUE;
END MkDgst;


END MD5.
